/*
 * Copyright (C) 2010-2018 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "SandboxExtension.h"
#include <WebCore/BlobPart.h>
#include <WebCore/Cookie.h>
#include <WebCore/FrameIdentifier.h>
#include <WebCore/NetworkLoadInformation.h>
#include <WebCore/PageIdentifier.h>
#include <WebCore/ResourceLoadStatistics.h>
#include <WebCore/ServiceWorkerTypes.h>
#include <pal/SessionID.h>
#include <utility>
#include <wtf/Forward.h>
#include <wtf/Optional.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace PAL {
class SessionID;
}

namespace WebCore {
class RegistrableDomain;
enum class StorageAccessWasGranted : bool;
struct SameSiteInfo;
enum class StorageAccessPromptWasShown : bool;
class ResourceError;
class NetworkLoadMetrics;
class ResourceRequest;
enum class IncludeSecureCookies : bool;
class ResourceResponse;
}

namespace WebKit {
class NetworkResourceLoadParameters;
class DownloadID;
}

namespace Messages {
namespace NetworkConnectionToWebProcess {

static inline IPC::StringReference messageReceiverName()
{
    return IPC::StringReference("NetworkConnectionToWebProcess");
}

class ScheduleResourceLoad {
public:
    typedef std::tuple<const WebKit::NetworkResourceLoadParameters&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ScheduleResourceLoad"); }
    static const bool isSync = false;

    explicit ScheduleResourceLoad(const WebKit::NetworkResourceLoadParameters& resourceLoadParameters)
        : m_arguments(resourceLoadParameters)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PerformSynchronousLoad {
public:
    typedef std::tuple<const WebKit::NetworkResourceLoadParameters&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("PerformSynchronousLoad"); }
    static const bool isSync = true;

    using DelayedReply = CompletionHandler<void(const WebCore::ResourceError& error, const WebCore::ResourceResponse& response, const Vector<char>& data)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::ResourceError& error, const WebCore::ResourceResponse& response, const Vector<char>& data);
    using Reply = std::tuple<WebCore::ResourceError&, WebCore::ResourceResponse&, Vector<char>&>;
    using ReplyArguments = std::tuple<WebCore::ResourceError, WebCore::ResourceResponse, Vector<char>>;
    explicit PerformSynchronousLoad(const WebKit::NetworkResourceLoadParameters& resourceLoadParameters)
        : m_arguments(resourceLoadParameters)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class TestProcessIncomingSyncMessagesWhenWaitingForSyncReply {
public:
    typedef std::tuple<const WebCore::PageIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("TestProcessIncomingSyncMessagesWhenWaitingForSyncReply"); }
    static const bool isSync = true;

    using DelayedReply = CompletionHandler<void(bool handled)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool handled);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit TestProcessIncomingSyncMessagesWhenWaitingForSyncReply(const WebCore::PageIdentifier& webPageID)
        : m_arguments(webPageID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class LoadPing {
public:
    typedef std::tuple<const WebKit::NetworkResourceLoadParameters&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("LoadPing"); }
    static const bool isSync = false;

    explicit LoadPing(const WebKit::NetworkResourceLoadParameters& resourceLoadParameters)
        : m_arguments(resourceLoadParameters)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveLoadIdentifier {
public:
    typedef std::tuple<uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RemoveLoadIdentifier"); }
    static const bool isSync = false;

    explicit RemoveLoadIdentifier(uint64_t resourceLoadIdentifier)
        : m_arguments(resourceLoadIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PageLoadCompleted {
public:
    typedef std::tuple<const WebCore::PageIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("PageLoadCompleted"); }
    static const bool isSync = false;

    explicit PageLoadCompleted(const WebCore::PageIdentifier& webPageID)
        : m_arguments(webPageID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PrefetchDNS {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("PrefetchDNS"); }
    static const bool isSync = false;

    explicit PrefetchDNS(const String& hostname)
        : m_arguments(hostname)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PreconnectTo {
public:
    typedef std::tuple<uint64_t, const WebKit::NetworkResourceLoadParameters&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("PreconnectTo"); }
    static const bool isSync = false;

    PreconnectTo(uint64_t preconnectionIdentifier, const WebKit::NetworkResourceLoadParameters& loadParameters)
        : m_arguments(preconnectionIdentifier, loadParameters)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StartDownload {
public:
    typedef std::tuple<const PAL::SessionID&, const WebKit::DownloadID&, const WebCore::ResourceRequest&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("StartDownload"); }
    static const bool isSync = false;

    StartDownload(const PAL::SessionID& sessionID, const WebKit::DownloadID& downloadID, const WebCore::ResourceRequest& request, const String& suggestedName)
        : m_arguments(sessionID, downloadID, request, suggestedName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ConvertMainResourceLoadToDownload {
public:
    typedef std::tuple<const PAL::SessionID&, uint64_t, const WebKit::DownloadID&, const WebCore::ResourceRequest&, const WebCore::ResourceResponse&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ConvertMainResourceLoadToDownload"); }
    static const bool isSync = false;

    ConvertMainResourceLoadToDownload(const PAL::SessionID& sessionID, uint64_t mainResourceLoadIdentifier, const WebKit::DownloadID& downloadID, const WebCore::ResourceRequest& request, const WebCore::ResourceResponse& response)
        : m_arguments(sessionID, mainResourceLoadIdentifier, downloadID, request, response)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CookiesForDOM {
public:
    typedef std::tuple<const PAL::SessionID&, const URL&, const WebCore::SameSiteInfo&, const URL&, const Optional<WebCore::FrameIdentifier>&, const Optional<WebCore::PageIdentifier>&, WebCore::IncludeSecureCookies> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("CookiesForDOM"); }
    static const bool isSync = true;

    using DelayedReply = CompletionHandler<void(const String& cookieString, bool didAccessSecureCookies)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& cookieString, bool didAccessSecureCookies);
    using Reply = std::tuple<String&, bool&>;
    using ReplyArguments = std::tuple<String, bool>;
    CookiesForDOM(const PAL::SessionID& sessionID, const URL& firstParty, const WebCore::SameSiteInfo& sameSiteInfo, const URL& url, const Optional<WebCore::FrameIdentifier>& frameID, const Optional<WebCore::PageIdentifier>& pageID, WebCore::IncludeSecureCookies includeSecureCookies)
        : m_arguments(sessionID, firstParty, sameSiteInfo, url, frameID, pageID, includeSecureCookies)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetCookiesFromDOM {
public:
    typedef std::tuple<const PAL::SessionID&, const URL&, const WebCore::SameSiteInfo&, const URL&, const Optional<WebCore::FrameIdentifier>&, const Optional<WebCore::PageIdentifier>&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetCookiesFromDOM"); }
    static const bool isSync = false;

    SetCookiesFromDOM(const PAL::SessionID& sessionID, const URL& firstParty, const WebCore::SameSiteInfo& sameSiteInfo, const URL& url, const Optional<WebCore::FrameIdentifier>& frameID, const Optional<WebCore::PageIdentifier>& pageID, const String& cookieString)
        : m_arguments(sessionID, firstParty, sameSiteInfo, url, frameID, pageID, cookieString)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CookiesEnabled {
public:
    typedef std::tuple<const PAL::SessionID&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("CookiesEnabled"); }
    static const bool isSync = true;

    using DelayedReply = CompletionHandler<void(bool enabled)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool enabled);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit CookiesEnabled(const PAL::SessionID& sessionID)
        : m_arguments(sessionID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CookieRequestHeaderFieldValue {
public:
    typedef std::tuple<const PAL::SessionID&, const URL&, const WebCore::SameSiteInfo&, const URL&, const Optional<WebCore::FrameIdentifier>&, const Optional<WebCore::PageIdentifier>&, WebCore::IncludeSecureCookies> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("CookieRequestHeaderFieldValue"); }
    static const bool isSync = true;

    using DelayedReply = CompletionHandler<void(const String& cookieString, bool didAccessSecureCookies)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& cookieString, bool didAccessSecureCookies);
    using Reply = std::tuple<String&, bool&>;
    using ReplyArguments = std::tuple<String, bool>;
    CookieRequestHeaderFieldValue(const PAL::SessionID& sessionID, const URL& firstParty, const WebCore::SameSiteInfo& sameSiteInfo, const URL& url, const Optional<WebCore::FrameIdentifier>& frameID, const Optional<WebCore::PageIdentifier>& pageID, WebCore::IncludeSecureCookies includeSecureCookies)
        : m_arguments(sessionID, firstParty, sameSiteInfo, url, frameID, pageID, includeSecureCookies)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetRawCookies {
public:
    typedef std::tuple<const PAL::SessionID&, const URL&, const WebCore::SameSiteInfo&, const URL&, const Optional<WebCore::FrameIdentifier>&, const Optional<WebCore::PageIdentifier>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetRawCookies"); }
    static const bool isSync = true;

    using DelayedReply = CompletionHandler<void(const Vector<WebCore::Cookie>& cookies)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<WebCore::Cookie>& cookies);
    using Reply = std::tuple<Vector<WebCore::Cookie>&>;
    using ReplyArguments = std::tuple<Vector<WebCore::Cookie>>;
    GetRawCookies(const PAL::SessionID& sessionID, const URL& firstParty, const WebCore::SameSiteInfo& sameSiteInfo, const URL& url, const Optional<WebCore::FrameIdentifier>& frameID, const Optional<WebCore::PageIdentifier>& pageID)
        : m_arguments(sessionID, firstParty, sameSiteInfo, url, frameID, pageID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteCookie {
public:
    typedef std::tuple<const PAL::SessionID&, const URL&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("DeleteCookie"); }
    static const bool isSync = false;

    DeleteCookie(const PAL::SessionID& sessionID, const URL& url, const String& cookieName)
        : m_arguments(sessionID, url, cookieName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RegisterFileBlobURL {
public:
    typedef std::tuple<const PAL::SessionID&, const URL&, const String&, const WebKit::SandboxExtension::Handle&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RegisterFileBlobURL"); }
    static const bool isSync = false;

    RegisterFileBlobURL(const PAL::SessionID& sessionID, const URL& url, const String& path, const WebKit::SandboxExtension::Handle& extensionHandle, const String& contentType)
        : m_arguments(sessionID, url, path, extensionHandle, contentType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RegisterBlobURL {
public:
    typedef std::tuple<const PAL::SessionID&, const URL&, const Vector<WebCore::BlobPart>&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RegisterBlobURL"); }
    static const bool isSync = false;

    RegisterBlobURL(const PAL::SessionID& sessionID, const URL& url, const Vector<WebCore::BlobPart>& blobParts, const String& contentType)
        : m_arguments(sessionID, url, blobParts, contentType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RegisterBlobURLFromURL {
public:
    typedef std::tuple<const PAL::SessionID&, const URL&, const URL&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RegisterBlobURLFromURL"); }
    static const bool isSync = false;

    RegisterBlobURLFromURL(const PAL::SessionID& sessionID, const URL& url, const URL& srcURL)
        : m_arguments(sessionID, url, srcURL)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RegisterBlobURLOptionallyFileBacked {
public:
    typedef std::tuple<const PAL::SessionID&, const URL&, const URL&, const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RegisterBlobURLOptionallyFileBacked"); }
    static const bool isSync = false;

    RegisterBlobURLOptionallyFileBacked(const PAL::SessionID& sessionID, const URL& url, const URL& srcURL, const String& fileBackedPath, const String& contentType)
        : m_arguments(sessionID, url, srcURL, fileBackedPath, contentType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RegisterBlobURLForSlice {
public:
    typedef std::tuple<const PAL::SessionID&, const URL&, const URL&, int64_t, int64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RegisterBlobURLForSlice"); }
    static const bool isSync = false;

    RegisterBlobURLForSlice(const PAL::SessionID& sessionID, const URL& url, const URL& srcURL, int64_t start, int64_t end)
        : m_arguments(sessionID, url, srcURL, start, end)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class UnregisterBlobURL {
public:
    typedef std::tuple<const PAL::SessionID&, const URL&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("UnregisterBlobURL"); }
    static const bool isSync = false;

    UnregisterBlobURL(const PAL::SessionID& sessionID, const URL& url)
        : m_arguments(sessionID, url)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class BlobSize {
public:
    typedef std::tuple<const PAL::SessionID&, const URL&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("BlobSize"); }
    static const bool isSync = true;

    using DelayedReply = CompletionHandler<void(uint64_t resultSize)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, uint64_t resultSize);
    using Reply = std::tuple<uint64_t&>;
    using ReplyArguments = std::tuple<uint64_t>;
    BlobSize(const PAL::SessionID& sessionID, const URL& url)
        : m_arguments(sessionID, url)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class WriteBlobsToTemporaryFiles {
public:
    typedef std::tuple<const PAL::SessionID&, const Vector<String>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WriteBlobsToTemporaryFiles"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Vector<String>&&)>&&);
    static void cancelReply(CompletionHandler<void(Vector<String>&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "WriteBlobsToTemporaryFilesReply" }; }
    using AsyncReply = CompletionHandler<void(const Vector<String>& fileNames)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<String>& fileNames);
    using Reply = std::tuple<Vector<String>&>;
    using ReplyArguments = std::tuple<Vector<String>>;
    WriteBlobsToTemporaryFiles(const PAL::SessionID& sessionID, const Vector<String>& blobURLs)
        : m_arguments(sessionID, blobURLs)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetCaptureExtraNetworkLoadMetricsEnabled {
public:
    typedef std::tuple<bool> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetCaptureExtraNetworkLoadMetricsEnabled"); }
    static const bool isSync = false;

    explicit SetCaptureExtraNetworkLoadMetricsEnabled(bool enabled)
        : m_arguments(enabled)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CreateSocketStream {
public:
    typedef std::tuple<const URL&, const PAL::SessionID&, const String&, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("CreateSocketStream"); }
    static const bool isSync = false;

    CreateSocketStream(const URL& url, const PAL::SessionID& sessionID, const String& cachePartition, uint64_t identifier)
        : m_arguments(url, sessionID, cachePartition, identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CreateSocketChannel {
public:
    typedef std::tuple<const PAL::SessionID&, const WebCore::ResourceRequest&, const String&, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("CreateSocketChannel"); }
    static const bool isSync = false;

    CreateSocketChannel(const PAL::SessionID& sessionID, const WebCore::ResourceRequest& request, const String& protocol, uint64_t identifier)
        : m_arguments(sessionID, request, protocol, identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class EnsureLegacyPrivateBrowsingSession {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("EnsureLegacyPrivateBrowsingSession"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(RESOURCE_LOAD_STATISTICS)
class RemoveStorageAccessForFrame {
public:
    typedef std::tuple<const PAL::SessionID&, const WebCore::FrameIdentifier&, const WebCore::PageIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RemoveStorageAccessForFrame"); }
    static const bool isSync = false;

    RemoveStorageAccessForFrame(const PAL::SessionID& sessionID, const WebCore::FrameIdentifier& frameID, const WebCore::PageIdentifier& pageID)
        : m_arguments(sessionID, frameID, pageID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)
class ClearPageSpecificDataForResourceLoadStatistics {
public:
    typedef std::tuple<const PAL::SessionID&, const WebCore::PageIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ClearPageSpecificDataForResourceLoadStatistics"); }
    static const bool isSync = false;

    ClearPageSpecificDataForResourceLoadStatistics(const PAL::SessionID& sessionID, const WebCore::PageIdentifier& pageID)
        : m_arguments(sessionID, pageID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)
class LogUserInteraction {
public:
    typedef std::tuple<const PAL::SessionID&, const WebCore::RegistrableDomain&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("LogUserInteraction"); }
    static const bool isSync = false;

    LogUserInteraction(const PAL::SessionID& sessionID, const WebCore::RegistrableDomain& domain)
        : m_arguments(sessionID, domain)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)
class ResourceLoadStatisticsUpdated {
public:
    typedef std::tuple<const Vector<std::pair<PAL::SessionID, Vector<WebCore::ResourceLoadStatistics>>>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ResourceLoadStatisticsUpdated"); }
    static const bool isSync = false;

    explicit ResourceLoadStatisticsUpdated(const Vector<std::pair<PAL::SessionID, Vector<WebCore::ResourceLoadStatistics>>>& statistics)
        : m_arguments(statistics)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)
class HasStorageAccess {
public:
    typedef std::tuple<const PAL::SessionID&, const WebCore::RegistrableDomain&, const WebCore::RegistrableDomain&, const WebCore::FrameIdentifier&, const WebCore::PageIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("HasStorageAccess"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "HasStorageAccessReply" }; }
    using AsyncReply = CompletionHandler<void(bool hasStorageAccess)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool hasStorageAccess);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    HasStorageAccess(const PAL::SessionID& sessionID, const WebCore::RegistrableDomain& subFrameDomain, const WebCore::RegistrableDomain& topFrameDomain, const WebCore::FrameIdentifier& frameID, const WebCore::PageIdentifier& pageID)
        : m_arguments(sessionID, subFrameDomain, topFrameDomain, frameID, pageID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)
class RequestStorageAccess {
public:
    typedef std::tuple<const PAL::SessionID&, const WebCore::RegistrableDomain&, const WebCore::RegistrableDomain&, const WebCore::FrameIdentifier&, const WebCore::PageIdentifier&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RequestStorageAccess"); }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebCore::StorageAccessWasGranted&&, WebCore::StorageAccessPromptWasShown&&)>&&);
    static void cancelReply(CompletionHandler<void(WebCore::StorageAccessWasGranted&&, WebCore::StorageAccessPromptWasShown&&)>&&);
    static IPC::StringReference asyncMessageReplyName() { return { "RequestStorageAccessReply" }; }
    using AsyncReply = CompletionHandler<void(WebCore::StorageAccessWasGranted wasGranted, WebCore::StorageAccessPromptWasShown promptWasShown)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, WebCore::StorageAccessWasGranted wasGranted, WebCore::StorageAccessPromptWasShown promptWasShown);
    using Reply = std::tuple<WebCore::StorageAccessWasGranted&, WebCore::StorageAccessPromptWasShown&>;
    using ReplyArguments = std::tuple<WebCore::StorageAccessWasGranted, WebCore::StorageAccessPromptWasShown>;
    RequestStorageAccess(const PAL::SessionID& sessionID, const WebCore::RegistrableDomain& subFrameDomain, const WebCore::RegistrableDomain& topFrameDomain, const WebCore::FrameIdentifier& frameID, const WebCore::PageIdentifier& pageID)
        : m_arguments(sessionID, subFrameDomain, topFrameDomain, frameID, pageID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)
class RequestStorageAccessUnderOpener {
public:
    typedef std::tuple<const PAL::SessionID&, const WebCore::RegistrableDomain&, const WebCore::PageIdentifier&, const WebCore::RegistrableDomain&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RequestStorageAccessUnderOpener"); }
    static const bool isSync = false;

    RequestStorageAccessUnderOpener(const PAL::SessionID& sessionID, const WebCore::RegistrableDomain& domainInNeedOfStorageAccess, const WebCore::PageIdentifier& openerPageID, const WebCore::RegistrableDomain& openerDomain)
        : m_arguments(sessionID, domainInNeedOfStorageAccess, openerPageID, openerDomain)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class AddOriginAccessWhitelistEntry {
public:
    typedef std::tuple<const String&, const String&, const String&, bool> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("AddOriginAccessWhitelistEntry"); }
    static const bool isSync = false;

    AddOriginAccessWhitelistEntry(const String& sourceOrigin, const String& destinationProtocol, const String& destinationHost, bool allowDestinationSubdomains)
        : m_arguments(sourceOrigin, destinationProtocol, destinationHost, allowDestinationSubdomains)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveOriginAccessWhitelistEntry {
public:
    typedef std::tuple<const String&, const String&, const String&, bool> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("RemoveOriginAccessWhitelistEntry"); }
    static const bool isSync = false;

    RemoveOriginAccessWhitelistEntry(const String& sourceOrigin, const String& destinationProtocol, const String& destinationHost, bool allowDestinationSubdomains)
        : m_arguments(sourceOrigin, destinationProtocol, destinationHost, allowDestinationSubdomains)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ResetOriginAccessWhitelists {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ResetOriginAccessWhitelists"); }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetNetworkLoadInformationRequest {
public:
    typedef std::tuple<uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetNetworkLoadInformationRequest"); }
    static const bool isSync = true;

    using DelayedReply = CompletionHandler<void(const WebCore::ResourceRequest& request)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::ResourceRequest& request);
    using Reply = std::tuple<WebCore::ResourceRequest&>;
    using ReplyArguments = std::tuple<WebCore::ResourceRequest>;
    explicit GetNetworkLoadInformationRequest(uint64_t resourceLoadIdentifier)
        : m_arguments(resourceLoadIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetNetworkLoadInformationResponse {
public:
    typedef std::tuple<uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetNetworkLoadInformationResponse"); }
    static const bool isSync = true;

    using DelayedReply = CompletionHandler<void(const WebCore::ResourceResponse& response)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::ResourceResponse& response);
    using Reply = std::tuple<WebCore::ResourceResponse&>;
    using ReplyArguments = std::tuple<WebCore::ResourceResponse>;
    explicit GetNetworkLoadInformationResponse(uint64_t resourceLoadIdentifier)
        : m_arguments(resourceLoadIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetNetworkLoadIntermediateInformation {
public:
    typedef std::tuple<uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetNetworkLoadIntermediateInformation"); }
    static const bool isSync = true;

    using DelayedReply = CompletionHandler<void(const Vector<WebCore::NetworkTransactionInformation>& transactions)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<WebCore::NetworkTransactionInformation>& transactions);
    using Reply = std::tuple<Vector<WebCore::NetworkTransactionInformation>&>;
    using ReplyArguments = std::tuple<Vector<WebCore::NetworkTransactionInformation>>;
    explicit GetNetworkLoadIntermediateInformation(uint64_t resourceLoadIdentifier)
        : m_arguments(resourceLoadIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class TakeNetworkLoadInformationMetrics {
public:
    typedef std::tuple<uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("TakeNetworkLoadInformationMetrics"); }
    static const bool isSync = true;

    using DelayedReply = CompletionHandler<void(const WebCore::NetworkLoadMetrics& networkMetrics)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::NetworkLoadMetrics& networkMetrics);
    using Reply = std::tuple<WebCore::NetworkLoadMetrics&>;
    using ReplyArguments = std::tuple<WebCore::NetworkLoadMetrics>;
    explicit TakeNetworkLoadInformationMetrics(uint64_t resourceLoadIdentifier)
        : m_arguments(resourceLoadIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(INDEXED_DATABASE)
class EstablishIDBConnectionToServer {
public:
    typedef std::tuple<const PAL::SessionID&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("EstablishIDBConnectionToServer"); }
    static const bool isSync = true;

    using DelayedReply = CompletionHandler<void(uint64_t serverConnectionIdentifier)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, uint64_t serverConnectionIdentifier);
    using Reply = std::tuple<uint64_t&>;
    using ReplyArguments = std::tuple<uint64_t>;
    explicit EstablishIDBConnectionToServer(const PAL::SessionID& sessionID)
        : m_arguments(sessionID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(SERVICE_WORKER)
class EstablishSWServerConnection {
public:
    typedef std::tuple<const PAL::SessionID&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("EstablishSWServerConnection"); }
    static const bool isSync = true;

    using DelayedReply = CompletionHandler<void(const WebCore::SWServerConnectionIdentifier& serverConnectionIdentifier)>;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::SWServerConnectionIdentifier& serverConnectionIdentifier);
    using Reply = std::tuple<WebCore::SWServerConnectionIdentifier&>;
    using ReplyArguments = std::tuple<WebCore::SWServerConnectionIdentifier>;
    explicit EstablishSWServerConnection(const PAL::SessionID& sessionID)
        : m_arguments(sessionID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

} // namespace NetworkConnectionToWebProcess
} // namespace Messages
